# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import string
import cProfile
import threading
import contextlib

try:
    import viztracer
except ImportError:
    viztracer = None

from contrast.utils.decorators import fail_loudly
from contrast.agent import request_state

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


VIZTRACER_SAVE_THREAD_NAME = "ContrastViztracerSave"
POSIX_SAFE_MAX_LENGTH = 64
POSIX_SAFE_CHARS = "_" + string.ascii_letters + string.digits

# VizTracer issues a stern warning if two tracers are recording simultaneously. We use
# this lock as non-blocking and only attempt to trace a request if some other request
# is not already being traced. This means we won't necessarily get viztracer output for
# every request, but it ensures that any data we do get is uncorrupted.
TRACER_LOCK = threading.Lock()


def _posix_safe(s: str) -> str:
    """
    Transform a string into a safe representation for a posix system filename. This
    function is overly aggressive but fine for our purposes.
    """
    s = s[:POSIX_SAFE_MAX_LENGTH].strip("/").replace("/", "_") or "_"
    return "".join(c for c in s if c in POSIX_SAFE_CHARS)


class Profiler(cProfile.Profile):
    def __init__(self, path: str):
        super().__init__()
        self.path = path

    def __enter__(self):
        self.enable()
        return self

    def __exit__(self, *exc_info):
        self.disable()
        self._save_profile_data()

    @fail_loudly("Unable to save profile data")
    def _save_profile_data(self):
        filename = (
            f"cprofile-{_posix_safe(self.path)}-{request_state.get_request_id()}.out"
        )
        logger.debug("writing cprofile data to %s", filename)
        self.dump_stats(filename)


@contextlib.contextmanager
def tracer(path: str, min_duration_ms: int = 0):
    if viztracer is not None:
        if TRACER_LOCK.acquire(blocking=False):
            min_duration_us = min_duration_ms * 1000
            try:
                request_tracer = viztracer.VizTracer(
                    verbose=0,
                    log_async=True,
                    min_duration=min_duration_us,
                )
                request_tracer.start()
                yield
                request_tracer.stop()
            finally:
                TRACER_LOCK.release()
            _save_trace_data(request_tracer, path)
            return
        else:
            logger.debug("Skipping viztracer for this request - already in use")
    else:
        logger.warning(
            "Attempted to enable tracing but couldn't import viztracer. "
            "Disable tracing or `pip install contrast-agent[debug]` extras."
        )
    yield


@fail_loudly("Unable to save trace data")
def _save_trace_data(request_tracer, path: str):
    filename = f"viztracer-{_posix_safe(path)}-{request_state.get_request_id()}.json.gz"
    logger.debug("writing viztracer data in a background thread", output_file=filename)
    threading.Thread(
        target=_do_save,
        args=(request_tracer, filename),
        name=VIZTRACER_SAVE_THREAD_NAME,
        daemon=True,
    ).start()


def _do_save(request_tracer, filename: str):
    request_tracer.save(output_file=filename)
    logger.debug("wrote viztracer data", output_file=filename)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.policy import propagation_policy
from contrast.agent.assess.policy.source_policy import apply_stream_source
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.propagators import STREAM_ACTIONS, stream_propagator
from contrast.agent.assess.utils import get_properties
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_and_patch_read_method(io_type, method_name):
    orig_func = getattr(io_type, method_name)
    if patch_manager.is_patched(orig_func):
        return

    propagation_method = STREAM_ACTIONS.get(method_name)

    def patched_func(wrapped, instance, args, kwargs):
        """
        This function has been optimized for performance. It looks similar to others in
        this module; however, do not attempt to deduplicate unless you're sure of the
        performance impact.
        """
        result = wrapped(*args, **kwargs)

        if scope.in_contrast_or_propagation_scope():
            return result

        with scope.propagation_scope():
            try:
                props = get_properties(instance)
                if props is not None and props.tags:
                    preshift = Preshift(instance, args, kwargs)
                    propagation_method(
                        method_name,
                        preshift,
                        result,
                        result,
                    )
                elif instance.cs__source:
                    apply_stream_source(
                        method_name, result, instance, result, args, kwargs
                    )
            except Exception as ex:
                _debug_log(method_name, ex)

        return result

    patch = wrap_and_watermark(orig_func, patched_func)
    patch_manager.patch(io_type, method_name, patch)


def build_and_patch_write_method(io_type, method_name):
    orig_func = getattr(io_type, method_name)
    if patch_manager.is_patched(orig_func):
        return

    def patched_func(wrapped, instance, args, kwargs):
        """
        This function has been optimized for performance. It looks similar to others in
        this module; however, do not attempt to deduplicate unless you're sure of the
        performance impact.
        """
        result = wrapped(*args, **kwargs)

        if scope.in_contrast_or_propagation_scope():
            return result

        with scope.propagation_scope():
            try:
                preshift = Preshift(instance, args, kwargs)
                stream_propagator.propagate_stream_write(
                    method_name, preshift, instance, result
                )
            except Exception as ex:
                _debug_log(method_name, ex)

        return result

    patch = wrap_and_watermark(orig_func, patched_func)
    patch_manager.patch(io_type, method_name, patch)


def build_and_patch_writelines(io_type, method_name):
    orig_func = getattr(io_type, method_name)
    if patch_manager.is_patched(orig_func):
        return

    def patched_func(wrapped, instance, args, kwargs):
        """
        This function has been optimized for performance. It looks similar to others in
        this module; however, do not attempt to deduplicate unless you're sure of the
        performance impact.
        """
        args_list = list(args[0:1]) + list(args[1:])

        result = wrapped(*args_list, **kwargs)

        if scope.in_contrast_or_propagation_scope():
            return result

        with scope.propagation_scope():
            try:
                propagation_policy.propagate_stream(
                    method_name,
                    result,
                    instance,
                    result,
                    args_list,
                    kwargs,
                )
            except Exception as ex:
                _debug_log(method_name, ex)

        return result

    patch = wrap_and_watermark(orig_func, patched_func)
    patch_manager.patch(io_type, method_name, patch)


def _debug_log(method_name, ex):
    logger.debug("failed to propagate %s", method_name, exc_info=ex)


def patch_getvalue(io_module):
    build_and_patch_read_method(io_module.StringIO, "getvalue")
    build_and_patch_read_method(io_module.BytesIO, "getvalue")


def patch_io(io_module):
    """
    Apply patches to methods of builtin stream types
    """
    for io_type in [io_module.StringIO, io_module.BytesIO]:
        build_and_patch_write_method(io_type, "write")

    # No need to hook StringIO.writelines because it is implemented as str.join under
    # the hood, so we already propagate for free. Unfortunately this might make the
    # reporting look a little odd, so we maybe should consider another solution later.
    build_and_patch_writelines(io_module.BytesIO, "writelines")

    patch_getvalue(io_module)


def register_patches():
    register_module_patcher(patch_io, "io")


def reverse_patches():
    unregister_module_patcher("io")
    io_module = sys.modules.get("io")
    if not io_module:
        return

    patch_manager.reverse_patches_by_owner(io_module)
    patch_manager.reverse_patches_by_owner(io_module.BytesIO)
    patch_manager.reverse_patches_by_owner(io_module.StringIO)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import sys

from contrast.agent.scope import contrast_scope, pop_contrast_scope
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.propagation_policy import propagate_string
from contrast_vendor import wrapt
from contrast.agent.assess.utils import get_properties
from contrast.utils.decorators import fail_quietly
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    get_arg,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)


class ContrastIteratorProxy(wrapt.ObjectProxy):
    def __init__(self, wrapped, path, _self_propagation_node):
        super().__init__(wrapped)

        try:
            # From os.fspath docs - If str or bytes is passed in, it is returned unchanged.
            # The point of this is to ensure we get a str/bytes object for the propagator.
            # path-like objects implement __fspath__ so we want the str/bytes value from that
            self._self_source_path = os.fspath(path)
        except TypeError:
            self._self_source_path = None

        self._self_propagation_node = _self_propagation_node

    def __enter__(self):
        self.__wrapped__.__enter__()
        return self

    def __iter__(self):
        return self

    @fail_quietly("Failed to propagate thru next iterator object")
    def _analyze_next(self, item):
        if item and self._self_source_path is not None:
            preshift = Preshift(None, [self._self_source_path], {})
            propagate_string(
                self._self_propagation_node,
                preshift,
                item.path,
                self.__wrapped__,
            )

    @contrast_scope()
    def __next__(self):
        item = None

        try:
            with pop_contrast_scope():
                item = self.__wrapped__.__next__()
        finally:
            self._analyze_next(item)
        return item


def build_os_scandir_hook(original_func, policy_node):
    """
    Builds patch for os.scandir(path="."). os.scandir returns an iterator of files/dir names based on the path argument
    """
    propagation_node = policy_node.propagator_nodes[0]

    @contrast_scope()
    def scandir_hook(wrapped, instance, args, kwargs):
        path = get_arg(args, kwargs, 0, "path")

        with pop_contrast_scope():
            dir_iter = wrapped(path)

        if get_properties(path) is None:
            return dir_iter

        dir_iter = ContrastIteratorProxy(dir_iter, path, propagation_node)

        return dir_iter

    return wrap_and_watermark(original_func, scandir_hook)


def patch_os(os_module):
    build_and_apply_patch(os_module, "scandir", build_os_scandir_hook)


def register_patches():
    register_module_patcher(patch_os, "os")


def reverse_patches():
    unregister_module_patcher("os")
    os_module = sys.modules.get("os")
    if not os_module:
        return

    patch_manager.reverse_patches_by_owner(os_module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import requests

from .base_ts_message import BaseTsAppMessage, b64url_stripped
from contrast import get_canonical_version
from contrast.utils.decorators import fail_loudly


class AgentStartup(BaseTsAppMessage):
    def __init__(self):
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"
        self.app_path_b64 = b64url_stripped(self.settings.app_path)

        self.body = {"version": get_canonical_version()}

        if code := self.settings.config.app_code:
            self.body["application_code"] = code

        if group := self.settings.config.app_group:
            self.body["application_group"] = group

        if metadata := self.settings.config.app_metadata:
            self.body["application_metadata"] = metadata

        if application_tags := self.settings.config.app_tags:
            self.body["application_tags"] = application_tags

        if environment := self.settings.config.get("server.environment"):
            self.body["server_environment"] = environment

        if server_tags := self.settings.config.get("server.tags"):
            self.body["server_tags"] = server_tags

        if session_id := self.settings.config.session_id:
            self.body["session_id"] = session_id
        elif session_metadata := self.settings.config.get_session_metadata():
            self.body["session_metadata"] = session_metadata

    @property
    def name(self) -> str:
        return "agent-startup"

    @property
    def path(self) -> str:
        return "/".join(
            [
                "agents",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                self.app_path_b64,
                "startup",
            ]
        )

    @property
    def expected_response_codes(self) -> list[int]:
        return [200]

    @property
    def request_method(self):
        return requests.put

    @property
    def disable_agent_on_401_and_408(self) -> bool:
        return True

    @fail_loudly("Failed to process AgentStartup response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()
        last_modified = response.headers.get("Last-Modified", None)
        # The agent was designed with the intention of having Server settings before Application settings.
        # This order must be maintained or a more significant refactor is required.
        self.settings.apply_server_settings(body.get("server_settings", None))
        self.settings.apply_application_settings(
            body.get("application_settings", None), last_modified
        )
        self.settings.apply_identification(body.get("identification", None))



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from contrast_fireball import InventoryComponent, ArchitectureComponent, Browser
import requests
from dataclasses import asdict

from contrast.agent.request import Request
from contrast.api.attack import Attack
from contrast.reporting.request_masker import RequestMasker

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ApplicationActivity(BaseTsAppMessage):
    def __init__(
        self,
        *,
        inventory_components: list[InventoryComponent] | None = None,
        attacks: list[Attack] | None = None,
        request: Request | None = None,
        request_data_masker: RequestMasker | None = None,
    ):
        super().__init__()

        if inventory_components is None:
            inventory_components = []
        if attacks is None:
            attacks = []

        self.body = {"lastUpdate": self.since_last_update}

        self.body["inventory"] = {
            # Used by TeamServer to aggregate counts across a given time period, for
            # Protect and attacker activity.
            "activityDuration": 0,
            "components": [],
        }
        if architecture_components := [
            asdict(c)
            for c in inventory_components
            if isinstance(c, ArchitectureComponent)
        ]:
            self.body["inventory"]["components"] = architecture_components
        if browsers := [c for c in inventory_components if isinstance(c, Browser)]:
            self.body["inventory"]["browsers"] = browsers
        if attacks and request:
            self.body["defend"] = {"attackers": []}

            for attack in attacks:
                self.body["defend"]["attackers"].append(
                    {
                        "protectionRules": {
                            attack.rule_id: attack.to_json(request, request_data_masker)
                        },
                        "source": {
                            "ip": request.client_addr or "",
                            "xForwardedFor": request.headers.get("X-Forwarded-For")
                            or "",
                        },
                    }
                )

    @property
    def name(self):
        return "application-activity"

    @property
    def path(self):
        return "activity/application"

    @property
    def request_method(self):
        return requests.put

    @property
    def expected_response_codes(self):
        return [200, 204]

    @fail_loudly("Failed to process Activity response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()

        self.settings.process_ts_reactions(body)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# ruff: noqa: F403

from .codecs import *
from .encodings import *
from .frameworks import *
from .glob import *
from .os import *
from .paths import *
from .re import *
from .serialize import *
from .string import *



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Generic utilities.
"""

from __future__ import annotations

import sys

from contextlib import suppress
from typing import Any


def get_processname() -> str:
    # based on code from
    # https://github.com/python/cpython/blob/313f92a57bc3887026ec16adb536bb2b7580ce47/Lib/logging/__init__.py#L342-L352
    processname = "n/a"
    mp: Any = sys.modules.get("multiprocessing")
    if mp is not None:
        # Errors may occur if multiprocessing has not finished loading
        # yet - e.g. if a custom import hook causes third-party code
        # to run when multiprocessing calls import.
        with suppress(Exception):
            processname = mp.current_process().name

    return processname



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

from contrast.utils.mapping import GlomDict


class ServerSettings:
    """
    This class is responsible for safely parsing V1 TeamServer Server Settings from a
    response to a usable format. The format can be found here:
    https://github.com/Contrast-Security-Inc/contrast-agent-api-spec

    At the time of the creation of this class, the specific schema is ServerSettings in
    agent-endpoints.yml.
    """

    def __init__(self, server_settings_json: dict | None = None):
        self.server_settings_json = GlomDict(server_settings_json or {})

    def common_config(self) -> dict[str, object]:
        """
        Returns the settings in a flattened common configuration format.
        """
        return {
            "server.environment": self.server_settings_json.get("environment", ""),
            "assess.enable": self.server_settings_json.get("assess.enable"),
            "assess.sampling.enable": self.server_settings_json.get(
                "assess.sampling.enable", False
            ),
            "assess.sampling.baseline": self.server_settings_json.get(
                "assess.sampling.baseline"
            ),
            "assess.sampling.request_frequency": self.server_settings_json.get(
                "assess.sampling.request_frequency"
            ),
            "assess.sampling.window_ms": (
                # teamserver incorrectly reports window_ms in seconds
                window * 1000
                if (
                    window := self.server_settings_json.get("assess.sampling.window_ms")
                )
                else None
            ),
            "agent.logger.level": self.server_settings_json.get("logger.level"),
            "agent.logger.path": self.server_settings_json.get("logger.path"),
            "agent.security_logger.syslog.enable": self.server_settings_json.get(
                "security_logger.syslog.enable"
            ),
            "agent.security_logger.syslog.protocol": self.server_settings_json.get(
                "security_logger.syslog.protocol"
            ),
            # security_logger.syslog.connection_type is ignored because the protocol
            # includes the connection type.
            "agent.security_logger.syslog.server_host": self.server_settings_json.get(
                "security_logger.syslog.ip"
            ),
            "agent.security_logger.syslog.port": self.server_settings_json.get(
                "security_logger.syslog.port"
            ),
            "agent.security_logger.syslog.facility": self.server_settings_json.get(
                "security_logger.syslog.facility"
            ),
            "agent.security_logger.syslog.severity_exploited": self.server_settings_json.get(
                "security_logger.syslog.severity_exploited"
            ),
            "agent.security_logger.syslog.severity_blocked": self.server_settings_json.get(
                "security_logger.syslog.severity_blocked"
            ),
            "agent.security_logger.syslog.severity_blocked_perimeter": self.server_settings_json.get(
                "security_logger.syslog.severity_blocked_perimeter"
            ),
            "agent.security_logger.syslog.severity_probed": self.server_settings_json.get(
                "security_logger.syslog.severity_probed"
            ),
            "agent.security_logger.syslog.severity_suspicious": self.server_settings_json.get(
                "security_logger.syslog.severity_suspicious"
            ),
            "protect.enable": self.server_settings_json.get("protect.enable"),
        }



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections.abc import Mapping
import functools

import contrast
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.assess.properties import Properties
from contrast.agent.assess.utils import (
    get_properties,
    get_last_event_from_sources,
    is_trackable,
    set_properties,
    track_string,
)
from contrast.agent.policy.constants import OBJECT, RETURN
from contrast.utils.assess.duck_utils import is_iterable, safe_getattr, safe_iterator
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils.safe_import import safe_import, safe_import_list
from contrast.utils.string_utils import truncated_signature

from tempfile import SpooledTemporaryFile

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Sentinel: ...


cgi_FieldStorage = safe_import("cgi.FieldStorage") or Sentinel


@functools.lru_cache(maxsize=1)
def _get_file_storage_types():
    return tuple(
        safe_import_list(
            "werkzeug.datastructures.FileStorage", "starlette.datastructures.UploadFile"
        )
    )


@functools.lru_cache(maxsize=1)
def _get_environ_headers_types():
    return tuple(safe_import_list("werkzeug.datastructures.EnvironHeaders"))


@functools.lru_cache(maxsize=1)
def _get_in_memory_uploaded_file_types():
    return tuple(
        safe_import_list("django.core.files.uploadedfile.InMemoryUploadedFile")
    )


@functools.lru_cache(maxsize=1)
def _get_file_field_types():
    return tuple(safe_import_list("aiohttp.web_request.FileField"))


def apply(nodes, self_obj, ret, orig_args, orig_kwargs=None, **kwargs):  # pylint: disable=redefined-builtin
    """
    Apply node into all sources to track all new strings coming in from a request
    """
    context = contrast.REQUEST_CONTEXT.get()
    if context is None or not context.assess_enabled:
        return

    file_storage_types = _get_file_storage_types()
    in_memory_uploaded_file_types = _get_in_memory_uploaded_file_types()
    for node in nodes:
        # args[0] is `self` for instance methods
        args = orig_args[1:] if node.instance_method else orig_args

        target = node.get_matching_first_target(self_obj, ret, args, orig_kwargs)

        source_type = node.type
        if isinstance(target, file_storage_types):
            # first, track the filename, if present
            filename = getattr(target, "filename", None)
            cs__apply_source(
                context,
                node,
                filename,
                target,
                filename,
                (),
                {},
                source_type="PARAMETER",
                source_name="filename",
            )
            # We don't try to propagate through werkzeug's file parsing logic, because
            # it causes massive slowdown. Instead, if we see a FileStorage object
            # come out of a multidict, we track it. We need to track the underlying
            # stream, not the FileStorage object itself.
            if hasattr(target, "stream"):
                # werkzeug case
                target = target.stream
            elif hasattr(target, "file"):
                # starlette case
                target = target.file
            source_type = "MULTIPART_CONTENT_DATA"
        elif isinstance(target, in_memory_uploaded_file_types):
            # same as above but for django
            source_type = "MULTIPART_CONTENT_DATA"
            if hasattr(target, "_name"):
                cs__apply_source(
                    context,
                    node,
                    target._name,
                    self_obj,
                    ret,
                    args,
                    orig_kwargs,
                    source_type=source_type,
                )
            target = target.file

        if hasattr(target, "cs__source"):
            # Target is already a source, nothing to do
            if target.cs__source:
                return

            target.cs__source = True
            target.cs__source_type = source_type
            target.cs__source_tags = node.tags
            set_properties(target, Properties(target))

        cs__apply_source(
            context,
            node,
            target,
            self_obj,
            ret,
            args,
            orig_kwargs,
            source_type=source_type,
        )


def adjust_source_tags(tags, source_type, source_name):
    """
    Update source tags to account for Referer header

    The XSS spec dictates that, with a single exception, no HTTP headers (or keys)
    should trigger XSS due to a high rate of false positives. The one exception is the
    'Referer' header. In order to account for this, we add the CROSS_SITE tag if the
    given source appears to correspond to this header value.
    """
    is_header = source_type == "HEADER"
    # This accounts for various representations of referer header keys, including
    # 'Referer' and 'HTTP_REFERER'
    is_referer = is_header and source_name and source_name.lower().endswith("referer")

    # According to XSS spec, the only header that should have CROSS_SITE is Referer.
    return (
        list(tags) + ["CROSS_SITE"]
        if (is_referer and "CROSS_SITE" not in tags)
        else list(tags)
    )


@fail_quietly("Error in apply_stream_source")
def apply_stream_source(method_name, target, self_obj, ret, args, kwargs):
    source_name = None
    context = contrast.REQUEST_CONTEXT.get()
    if context is None or not context.assess_enabled:
        return

    module = self_obj.__class__.__module__
    class_name = self_obj.__class__.__name__

    source_type = safe_getattr(self_obj, "cs__source_type", None) or "BODY"
    source_tags = safe_getattr(self_obj, "cs__source_tags", None) or ["UNTRUSTED"]

    if context.stop_source_creation(source_type, source_name):
        return

    node = SourceNode(
        module,
        class_name,
        True,
        method_name,
        "RETURN",
        source_type,
        tags=source_tags,
    )

    cs__apply_source(context, node, target, self_obj, ret, args, kwargs)


NAMED_SOURCE_TYPES = [
    "PARAMETER",
    "PARAMETER_KEY",
    "HEADER",
    "HEADER_KEY",
    "COOKIE",
    "COOKIE_KEY",
    "MULTIPART_HEADER",
    "MULTIPART_HEADER_KEY",
    "MULTIPART_PARAMETER",
    "MULTIPART_PARAMETER_KEY",
]


def _get_source_name(node, self_obj, args):
    """
    Determine name for source

    Some sources require a source name since there can potentially be multiple values
    for that given source type (e.g. multiple headers, cookies, etc.).

    If no source name is explicitly given, but we have a dict/mapping object, then we
    try to use the dict/mapping key as the source name. This assumes that the
    instrumented dict/mapping method is something like __getitem__.

    Otherwise, use the method name (e.g. for `flask.wrappers.Request.full_path`, the
    source name will be "full_path").
    """
    if args and isinstance(self_obj, (dict, Mapping) + _get_environ_headers_types()):
        return args[0]

    return node.method_name


def cs__apply_source(
    context,
    node,
    target,
    self_obj,
    ret,
    args,
    kwargs,
    source_name=None,
    source_type=None,
):
    logger.debug("In cs__apply_source for %r", node)

    if source_type is None:
        source_type = node.type

    if source_type in NAMED_SOURCE_TYPES and not source_name:
        source_name = _get_source_name(node, self_obj, args)

    if context.stop_source_creation(source_type, source_name):
        return

    _cs__apply_source(
        context,
        node,
        target,
        self_obj,
        ret,
        source_type,
        source_name,
        True,
        args,
        kwargs,
    )


@fail_loudly("Unable to apply source for node")
def _cs__apply_source(
    context,
    node,
    target,
    self_obj,
    ret,
    source_type,
    source_name,
    use_source_key,
    args,
    kwargs,
):
    if not context or not node or target is None:
        return

    if context.stop_source_creation(source_type, source_name):
        return

    target_properties = get_properties(target)
    if target_properties is not None and all(
        e.source_type == "COOKIE" for e in target_properties.events if e.source_type
    ):
        source_type = "COOKIE"
        # Discard any preexisting tags or events for this string. This greatly
        # simplifies reporting by retaining only the most recent source event for each
        # string
        target_properties.reset()
    if target_properties is None and is_trackable(target):
        target_properties = track_string(target)
    if target_properties is not None:
        if not safe_getattr(target, "cs__source", False):
            for tag in adjust_source_tags(node.tags, source_type, source_name):
                if source_type == "COOKIE" and tag == "CROSS_SITE":
                    continue

                target_properties.add_tag(tag, AdjustedSpan(0, len(target)))
                target_properties.add_properties(node.properties)

                logger.debug(
                    "Source %s detected: %s tagged with %s",
                    node.name,
                    truncated_signature(target),
                    str(tag),
                )

        parents = get_parents(node, self_obj, ret, args, kwargs)
        target_properties.build_event(
            node,
            target,
            self_obj,
            ret,
            args,
            kwargs,
            parents,
            possible_key=None,
            source_type=source_type,
            source_name=source_name,
        )

        # Only report the source if we have written policy for it.
        # If we were to report all sources, then events such as pre-tracking
        # the WSGI environ would result in a lot of noise. (PYT-825)
        if node.policy_patch:
            target_properties.report_source(target_properties.event)

        context.source_created()

        if is_trackable(target):
            set_properties(target, target_properties)

    elif isinstance(target, (dict, Mapping)):
        # This logic is intended to handle the case where the source is a dict
        # that happens to represent request parameters, headers, or cookies.
        # In these cases we want the top-level keys of this dictionary to be
        # labeled with the appropriate data type (e.g. PARAMETER_KEY). This
        # behavior should not apply to recursive calls, however.
        source_key_type = key_type(source_type) if use_source_key else source_type
        for key, value in target.items():
            _cs__apply_source(
                context,
                node,
                key,
                self_obj,
                ret,
                source_key_type,
                key,
                False,
                args,
                kwargs,
            )
            _cs__apply_source(
                context,
                node,
                value,
                self_obj,
                ret,
                source_type,
                key,
                False,
                args,
                kwargs,
            )

    elif isinstance(target, cgi_FieldStorage):
        # we have patches for this already and we don't want to fall into the
        # is_iterable case
        return

    elif isinstance(target, SpooledTemporaryFile):
        # circular import
        from contrast.utils.assess.stream_utils import ContrastFileProxy  # noqa

        target._file = ContrastFileProxy(target._file)
        target._file.cs__source_type = "MULTIPART_CONTENT_DATA"
        target._file.cs__source_tags = node.tags

    elif isinstance(target, _get_file_field_types()):
        # circular import
        from contrast.utils.assess.stream_utils import (
            ContrastBufferedReaderProxy,
        )  # noqa

        # these instances are frozen using a library called `attr`. The library
        # overrides __setattr__ for instances. The documented workaround is to use
        # `object`'s __setattr__ instead
        object.__setattr__(target, "file", ContrastBufferedReaderProxy(target.file))
        target.file.cs__source_type = "MULTIPART_CONTENT_DATA"
        target.file.cs__source_tags = node.tags

    elif is_iterable(target):
        for value in safe_iterator(target):
            _cs__apply_source(
                context,
                node,
                value,
                self_obj,
                ret,
                source_type,
                source_name,
                False,
                args,
                kwargs,
            )


def value_of_source(source, self_obj, ret, args, kwargs):
    if source == OBJECT:
        return self_obj

    if source == RETURN:
        return ret

    if not args:
        return self_obj

    if args and isinstance(source, int) and source < len(args):
        return args[source]

    if kwargs and source in kwargs:
        return kwargs[source]

    return None


def get_parents(node, self_obj, ret, args, kwargs):
    if isinstance(node, SourceNode):
        return []

    sources = [
        value_of_source(source, self_obj, ret, args, kwargs) for source in node.sources
    ]
    return get_last_event_from_sources(sources)


def build_source_node(source_dict, method_name, source_type):
    """
    Useful for explicit source tracking, ie scope tracking / aiohttp request tracking.

    Note that headers (including cookies) should not trigger XSS, per labs
    """
    tags = (
        ["CROSS_SITE"] if source_type not in ["HEADER", "COOKIE"] else ["NO_NEWLINES"]
    )

    node = SourceNode(
        method_name=method_name,
        node_type=source_type,
        tags=tags,
        **source_dict,
    )
    node.skip_stacktrace = True

    return node


KEY_MAPPING = {
    "PARAMETER": "PARAMETER_KEY",
    "HEADER": "HEADER_KEY",
    "COOKIE": "COOKIE_KEY",
}

PARAMETER_TYPE = "PARAMETER"
PARAMETER_KEY_TYPE = "PARAMETER_KEY"
HEADER_TYPE = "HEADER"
HEADER_KEY_TYPE = "HEADER_KEY"
COOKIE_TYPE = "COOKIE"
COOKIE_KEY_TYPE = "COOKIE_KEY"
QUERYSTRING_TYPE = "QUERYSTRING"


def key_type(source_type):
    return KEY_MAPPING.get(source_type, source_type)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.settings import Settings
from contrast.utils.digest_utils import Digest
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class BaseRule:
    """
    Base rule object that all assess rules will inherit
    """

    def __init__(self):
        self.findings_per_agent_run = 0

    @property
    def name(self):
        raise NotImplementedError

    @property
    def disabled(self):
        """
        Property indicating whether rule is disabled
        """
        return Settings().is_assess_rule_disabled(self.name)

    def finding_reported(self):
        self.findings_per_agent_run += 1

    def generate_preflight_hash(self, **kwargs):
        hasher = Digest()
        hasher.update(self.name)

        self.update_preflight_hash(hasher, **kwargs)

        return hasher.finish()

    def update_preflight_hash(self, hasher, **kwargs):
        """
        Update preflight hash with additional rule-specific data

        Child classes should override this method in order to customize the
        kind of data that is used to generate the preflight hash.

        @param hasher: Hash class to be updated with additional data
        @param **kwargs: Placeholder for keyword args used by child classes
        """
        raise NotImplementedError

    def count_threshold_reached(self):
        """
        Use the configurable max vulnerability threshold and agent runtime window
        to determine if we've already created the maximum number of
        vulnerabilities for this rule.

        RACE CONDITION:  a possible race condition can happen of multiple threads
        check this method. The outcome is relatively low risk - worst case we will
        end up making more vulnerabilities than the threshold.
        """
        settings = Settings()
        result = True

        if self.findings_per_agent_run + 1 <= settings.max_vulnerability_count:
            result = False
        else:
            logger.warning(
                "%s vulnerability will not be reported. %s vulnerabilities already reported",
                self.name,
                self.findings_per_agent_run,
            )

        if now_ms() - settings.agent_runtime_window >= settings.agent_runtime_threshold:
            self.findings_per_agent_run = 0
            settings.agent_runtime_window = now_ms()

        return result



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.agent.middlewares.response_wrappers.base_response_wrapper import (
    BaseResponseWrapper,
)


def get_otel_attributes(response: BaseResponseWrapper) -> dict:
    """
    Returns attributes following OpenTelemetry semantic conventions for HTTP responses.
    """
    attributes = {
        "http.response.status_code": response.status_code,
    }
    if response.status_code >= 500:
        attributes["error.type"] = str(response.status_code)

    if response.headers and (content_type := response.headers.get("Content-Type")):
        attributes["http.response.header.content_type"] = content_type

    return attributes



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.base_rule import BaseRule
from contrast.agent.assess.rules import build_finding, send_finding


class StaticRule(BaseRule):
    """
    Base class for provider and config-based rules

    These kinds of rules send their findings immediately rather than relying on a
    request context.
    """

    def build_and_send_finding(self, properties, **kwargs):
        finding = build_finding(self, properties, **kwargs)

        self.finding_reported()
        send_finding(finding)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import contrast
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.utils.assess.duck_utils import safe_getattr
from contrast.utils.assess.tag_utils import merge_tags


def get_properties(value):
    return safe_getattr(value, "cs__properties", None)


def set_properties(obj, props):
    obj.cs__properties = props


def clear_properties():
    contrast.STRING_TRACKER.clear()


def is_trackable(obj):
    return hasattr(obj, "cs__properties")


def is_tracked(value):
    return value in contrast.STRING_TRACKER


def track_string(value):
    return contrast.STRING_TRACKER.track(value)


def track_and_tag(value, label="UNTRUSTED", start=None, end=None):
    properties = track_string(value)
    if not properties:
        return None

    if start is None:
        start = 0
    if end is None:
        end = len(value)

    properties.add_tag(label, AdjustedSpan(start, end))
    return properties


def copy_events(target_props, source_props):
    if (
        source_props is None
        or target_props is None
        or target_props is source_props
        or source_props.event is None
    ):
        return

    target_props.event = source_props.event


def copy_from(to_obj, from_obj, shift=0, skip_tags=None):
    """Copy events and tags from from_obj to to_obj"""
    if from_obj is to_obj:
        return

    if (from_props := get_properties(from_obj)) is None:
        return

    # we assume to_obj has already been tracked and has properties
    to_props = get_properties(to_obj)

    if from_props == to_props:
        return

    copy_events(to_props, from_props)

    for key in from_props.tag_keys():
        if skip_tags and key in skip_tags:
            continue

        from_props_tags = from_props.fetch_tags(key)
        new_tags = [tag.copy_modified(shift) for tag in from_props_tags]

        existing_tags = to_props.fetch_tags(key)

        if existing_tags:
            existing_tags.extend(new_tags)
        else:
            to_props.set_tag(key, new_tags)


def get_last_event_from_sources(sources):
    """
    Gathers from given sources the parent IDs that should be used for an event
    """
    return [
        props.event
        for source in sources
        if (props := get_properties(source)) and props.event
    ]


def copy_tags_in_span(target, source_properties, span, offset=0):
    """
    Given source properties, copies tags at a given span to the target
    """
    span = AdjustedSpan(*tuple(x + offset for x in span))
    source_tags = source_properties.tags_at_range(span)
    if not source_tags:
        return get_properties(target)

    target_properties = track_string(target)
    if target_properties is None:
        return get_properties(target)

    for name, tags in source_tags.items():
        for tag in tags:
            target_properties.add_existing_tag(name, tag)

    merge_tags(target_properties.tags)
    return target_properties


def copy_tags_to_offset(target_properties, source_tags, target_offset):
    """
    Given source tags, copy to target properties at offset.

    The caller is responsible for updating the string tracker if necessary.
    """
    for name, tags in source_tags.items():
        for tag in tags:
            new_tag = tag.copy_modified(target_offset)
            target_properties.add_existing_tag(name, new_tag)
